/**
 * 
 */
package org.prelle.shell;

import java.io.InputStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import org.apache.log4j.Logger;
import org.prelle.config.commands.ActionResult;
import org.prelle.config.commands.Command;
import org.prelle.config.commands.DataColumn;
import org.prelle.config.commands.DataRow;
import org.prelle.config.commands.DataSet;
import org.prelle.user.ConfigUser;

/**
 * @author prelle
 *
 */
public class BaseShell {

	public final static char INDICATE_TAB = (char)9;

	private Logger logger = Logger.getLogger("shell");

	private BaseShellListener listener;
	private List<ConnectedUser> user;

	//------------------------------------------------------
	/**
	 */
	public BaseShell(BaseShellListener list) {
		this.listener = list;
		user = new ArrayList<ConnectedUser>();
	}

	//------------------------------------------------------
	public ConnectedUser createNewUserConnection(InputStream in, PrintWriter out) {
		ConnectedUser ret = new ConnectedUser(in,out, null);
		user.add(ret);
		return ret;
	}

	//------------------------------------------------------
	public Collection<ConnectedUser> getConnectedUser() {
		List<ConnectedUser> ret = new ArrayList<ConnectedUser>(user);
		return ret;
	}

	//------------------------------------------------------
	public void removeUser(ConnectedUser toDelete) {
		user.remove(toDelete);
	}

	//------------------------------------------------------
	/**
	 * @param dir
	 * @param accessLevel
	 * @return
	 */
	public String getPrompt(CommandContext dir) {
		if (dir==null)
			return "(null)> ";
		return dir.toPrompt()+"> ";
	}

	//--------------------------------------------------------------
	public void parseLoggedIn(ConnectedUser user, String line) {
		if (line!=null && line.length()>0) {
			logger.debug("----------Parsing: \""+line+"\"");
			boolean isHelp = line.charAt(line.length()-1)=='?';
			boolean isTab  = line.charAt(line.length()-1)==INDICATE_TAB;
			logger.debug("isHelp="+isHelp+"   isTab="+isTab);
			String[] reply = null;
			if (isHelp) {
				String line2 = line.substring(0, line.length()-1);
				logger.debug("----------Helping: \""+line2+"\"");
				reply = perform(user, line2, true);
			} else if (isTab){
				String line2 = line.substring(0, line.length()-1);
				logger.debug("----------Tabbing: \""+line2+"\"");
				String complete = complete(user, line2);
				logger.debug("Complete:["+complete+"]");
				if  (complete==null || complete.length()==0) {
					return ;
				}
				user.getOut().print(complete);
				//				reply = new String[]{"",getPrompt(user.getDir())+complete};
				user.buffer.setInput(line.substring(0,line.length()-1)+complete);
				user.getOut().flush();
				return;
			} else {
				reply = perform(user, line, false);
			}
			user.remember = reply[1];
			user.getOut().print(reply[0]+user.remember);
			user.getOut().flush();
		} else {
			user.getOut().print(getPrompt(user.getDir()));
			user.getOut().flush();
		}
	}

	//------------------------------------------------------
	private String complete(ConnectedUser user, String line) {
		CommandContext context = user.getDir();
		List<Command> com = context.getPossibleCommandCompletions(line);
		switch (com.size()) {
		case 0:
			// Play bell
			return null;
		case 1:
			// Only one possibility - send missing characters
			String cname = com.get(0).getCommandName();
			return cname.substring(line.length(), cname.length())+" ";
		default:			
			return line;
		}
	}

	//--------------------------------------------------------------
	private String[] perform(ConnectedUser user, String line, boolean help) {
		logger.info("perform '"+line+"' at "+user.getDir());
		PushbackStringTokenizer tok = new PushbackStringTokenizer(line, " \r\n");

		String repeat  = help?line:"";
		String message = perform(user.getDir(), user, tok, help);
		String prompt  = getPrompt(user.getDir());
		if (prompt==null) {
			logger.debug("No prompt found");
			prompt = "> ";
		}

		if (message==null) 
			return new String[]{prompt, repeat};
		else
			return new String[]{message+"\n\r"+prompt, repeat};

	}

	//--------------------------------------------------------------
	public String perform(CommandContext path, ConnectedUser user, PushbackStringTokenizer tok, boolean wasHelp) {
		logger.debug("perform: at "+path.toPrompt()+" / wasHelp="+wasHelp+", input="+tok);
		String command = tok.nextToken();
		if (command==null || command.length()==0) {
			logger.debug("No command entered");
			/*
			 * List all available commands
			 */
			StringBuffer ret = new StringBuffer();
			for (Command com : path.getAvailableCommands()) {
				ret.append(com.getCommandName()+" \t- "+com.getShortHelp()+"\r\n");
			}
			return ret.toString();
		}

		logger.debug("Command = "+command);

		StringBuffer ret = new StringBuffer();
		if (command.equalsIgnoreCase("help")) {
			ret.append("----------------------------------------------\r\n");
			wasHelp = true;
			if (!tok.hasMoreTokens()) {
				/*
				 * Help has simply been called without additional parameters
				 */
				// For command trees
				if (path instanceof TreeCommandContext) {
					logger.debug("add description");
					ret.append( ((TreeCommandContext)path).getDescription()+"\r\n");
				}
				if (path.getAvailableCommands().size()==0)
					ret.append("No commands available\r\n");
				for (Command com : path.getAvailableCommands()) {
					ret.append(com.getCommandName()+" \t- "+com.getShortHelp()+"\r\n");
				}

				return ret.toString();
			}
			// Help with additional parameter
			command = tok.nextToken();
		} 

		// Parse into tokens
		ArrayList<Object> tokens = new ArrayList<Object>();
//		tokens.add(command);
		for (String token : tok.getTokenList()) {
			if (token.equalsIgnoreCase("On") || token.equalsIgnoreCase("true")  || token.equalsIgnoreCase("enable") )
				tokens.add(true);
			else if (token.equalsIgnoreCase("Off") || token.equalsIgnoreCase("false")  || token.equalsIgnoreCase("disable") )
				tokens.add(false);
			else 
				try {
					tokens.add(Integer.parseInt(token));
				} catch (NumberFormatException nfe) {
					tokens.add(token);
				}
		}
		logger.debug("Tokens = "+tokens);
		Object[] input = new Object[tokens.size()];
		input = tokens.toArray(input);

		/*
		 * Search the command to execute
		 */
		StringBuffer output = new StringBuffer();
		
		for (Command com : path.getAvailableCommands()) {
			if (com.getCommandName().startsWith(command)) {
				logger.info("Execute command "+com+" using"+tokens);
				if (wasHelp) {
					if (input.length>0)
						return com.getLongHelp(""+input[1]);
					return com.getLongHelp();
				}
				logger.debug("tok = "+tok);
				logger.debug("tok.left = "+tok.getTokensLeft());
				logger.debug("tok.list = "+tok.getTokenList());
				logger.debug("tokens   = "+tokens);
				logger.debug("input    = "+Arrays.toString(input));
				
				logger.info("Execute "+com.getClass());
				ActionResult result = com.execute(path, input);
				// Sanity check
				logger.info("executing resulted in "+result);
				if (result==null) {
					logger.warn("Executing "+com.getClass()+" resulted in null");
					return "Error executing command";
				}
				// process result
				switch (result.getType()) {
				case ActionResult.NO_FEEDBACK:
					return null;
				case ActionResult.DISPLAY_TEXT: 
					output.append( result.getDataAsText() +"\n");
					if (result.hasSuccessor()) {
						logger.debug("ActionResult has successor");
						ActionResult result2 = result.getSuccessor();
						switch (result.getType()) {
						case ActionResult.DISPLAY_TEXT: 
							output.append( result2.getDataAsText());
							break;
						case ActionResult.DISPLAY_LIST:
							output.append( generateList(result2.getDataAsList(), user.getTerminalWidth()));
							break;
						}
					}
					return output.toString();
				case ActionResult.DISPLAY_LIST:
					output.append( generateList(result.getDataAsList(), user.getTerminalWidth()) );
					if (result.hasSuccessor()) {
						logger.debug("ActionResult has successor");
						ActionResult result2 = result.getSuccessor();
						switch (result.getType()) {
						case ActionResult.DISPLAY_TEXT: 
							output.append( result2.getDataAsText());
							break;
						case ActionResult.DISPLAY_LIST:
							output.append( generateList(result2.getDataAsList(), user.getTerminalWidth()));
							break;
						}
					}
					return output.toString();
				case ActionResult.LOGOUT:
					logger.info("User "+user+" disconnected");
					user.disconnect();
					return null;
				case ActionResult.CHANGE_DIR:
					logger.info("User "+user+" changes to "+result.getChangeTo());
					if (result.getChangeTo()!=null) {
						user.setDir((CommandContext) result.getChangeTo());
						path = (CommandContext) result.getChangeTo();
					}
					// Interpret parameters as new command input
					logger.debug("Remaining tokens: "+tok.getTokenList());
					if (tok.hasMoreTokens())
						return perform(path, user, tok, wasHelp);
					return null;
				}
				logger.warn("ActionResult type not implemented");
				return "ActionResult type not implemented";
			}
		}


		return "No such command '"+command+"'";
	}

	//------------------------------------------------------
	/**
	 * @return
	 */
	public CommandContext getInitialContext(ConfigUser user) {
		logger.warn("getInitialContext for "+user.getName()+" = "+listener.getStartPath(user));
		return listener.getStartPath(user);
	}

	//------------------------------------------------------
	@SuppressWarnings("unchecked")
	private String generateList(DataSet set, int maxWidth) {
//		// Show only visible or ID columns
//		ArrayList<DataColumn> foo = new ArrayList<DataColumn>();
//		for (DataColumn col : set.getColumns())
//			if (col.isVisible() || col.isID())
//				foo.add(col);
//		DataColumn[] columns = new DataColumn[foo.size()];
//		columns = foo.toArray(columns);
		DataColumn[] columns = set.getColumns();
		
		int cols = columns.length;
		ArrayList<Integer>[] widths = new ArrayList[cols];
		int[] minWidths = new int[cols];
		int[] max = new int[cols];
		int[] sum = new int[cols];

		// Column headings as first indicator for width
		for (int col=0; col<cols; col++) {
			DataColumn column = columns[col];
			if (!column.isVisible() && !column.isID())
				continue;
			max[col] = columns[col].getHeading().length();
			//			sum[col] = max[col];
			minWidths[col] = max[col];
			widths[col] = new ArrayList<Integer>();
		}
		
		Iterator<DataRow> rows = set.iterator();
		while (rows.hasNext()) {
			DataRow row = rows.next();
			for (int col=0; col<cols; col++) {
				int len = (row.getData()[col]+"").length();
				sum[col] += len;
				if (len>max[col])
					max[col] = len;
				if (widths[col]==null) widths[col] = new ArrayList<Integer>();
				widths[col].add(len);
				logger.debug(row.getData()[col]+" read - width now "+max[col]);
			}
		}

		/*
		 * By now sum[] and max[] regard all values
		 * Now try to see if max-width is possible, otherwise
		 * use average width
		 * Always regard that per column +2 must be added 
		 * and +2 for the final table border
		 */

		// What would be the maximal width
		int maxSum = 1;
		for (int col=0; col<cols; col++) {
			maxSum += max[col]+3;
		}
		logger.debug("Maximal width would be "+maxSum);

		int[] avgWidth = new int[cols];
		int avgSum = 1;
		for (int col=0; col<cols; col++) {
			logger.debug("---sum[col]="+sum[col]);
			logger.debug("---widths[col]="+widths[col]);
			avgWidth[col] = Math.round( (float)sum[col] / (float)widths[col].size());
			avgSum += avgWidth[col]+3;
			logger.debug("avgWidth of Col "+col+" = "+avgWidth[col]+"     avgSum = "+avgSum);

			// Ensure minimim widths to display header
			if (avgWidth[col]<minWidths[col])
				avgWidth[col] = minWidths[col];
		}
		logger.debug("Average width would be "+avgSum);

		int[] useWidth = null;
		logger.debug("maxSum="+maxSum+"    maxWidth="+maxWidth);
		if (maxSum<=maxWidth) {
			useWidth = max;
			logger.debug("Use maximal width for each column");
		} else {
			useWidth = avgWidth;
			logger.debug("Use average width for each column");
			/*
			 * distribute available space
			 */
			int left = maxWidth - avgSum;
			while (left>0) {
				for (int col=0; col<cols; col++) {
					if (avgWidth[col]<max[col]) {
						avgWidth[col]++;
						left--;
						if (left==0)
							break;
					}
				}
			}
		}

		// Create a seperator
		StringBuffer sep = new StringBuffer();
		for (int col=0; col<cols; col++) {
			sep.append("+-");
			for (int i=0; i<useWidth[col]; i++)
				sep.append('-');
			sep.append('-');
		}
		sep.append("+");

		/*
		 * Build a format line
		 */
		StringBuffer formatLine = new StringBuffer();
//		DataColumn[] columns = set.getColumns();
		for (int col=0; col<cols; col++) {
			formatLine.append("| %"+(col+1)+"$");
			if (columns[col].getType()==DataColumn.TYPE_STRING)
				formatLine.append("-");
			formatLine.append((useWidth[col])+".");
			formatLine.append((useWidth[col])+"s");
			formatLine.append(" ");
		}
		formatLine.append("|");
		logger.debug("Format pattern is "+formatLine);

		/*
		 * Generate output
		 */
		StringBuffer out = new StringBuffer();
		// Table start
		out.append(sep.toString()+"\r\n");
		// Heading
		Object[] heads = new String[cols];
		for (int col=0; col<cols; col++)
			heads[col]=columns[col].getHeading();
		out.append(String.format(formatLine.toString(), heads));
		out.append("\r\n");

		// Heading end
		out.append(sep.toString()+"\r\n");
		// Content
		rows = set.iterator();
		while (rows.hasNext()) {
			DataRow row = rows.next();
			Object[] values = row.getData();
			out.append(String.format(formatLine.toString(), values));
			out.append("\r\n");
		}
		// Table end
		out.append(sep.toString()+"\r\n");

		return out.toString();
	}

}
